@@ -351,7 +351,8 @@ int rcksum_submit_source_data(struct rcksum_state *const z, unsigned char *data,
                 /* Do a hash table lookup - first in the bithash (fast negative
                  * check) and then in the rsum hash */
                 unsigned hash = z->r[0].b;
-                hash ^= ((z->seq_matches > 1) ? z->r[1].b : z->r[0].a) << BITHASHBITS;
+                hash ^= ((z->seq_matches > 1) ? z->r[1].b
+                        : z->r[0].a & z->rsum_a_mask) << BITHASHBITS;
                 if ((z->bithash[(hash & z->bithashmask) >> 3] & (1 << (hash & 7))) != 0
                     && (e = z->rsum_hash[hash & z->hashmask]) != NULL) {
 
@@ -436,10 +437,6 @@ int rcksum_submit_source_file(struct rcksum_state *z, FILE * f, int progress) {
         /* If this is the start, fill the buffer for the first time */
         if (!in) {
             len = fread(buf, 1, bufsize, f);
-            if (len < z->context) {
-                free(buf);
-                return 0;
-            }
             in += len;
         }
 
